# Histograms.
#
# Written by Konrad Hinsen <hinsen@cnrs-orleans.fr>
# last revision: 2001-11-30
#

import Numeric; N = Numeric

class Histogram:

    """Histogram in one variable

    Constructor: Histogram(|data|, |bins|, |range|=None)

    Arguments:

    |data| -- a sequence of data points

    |bins| -- the number of bins into which the data is to be sorted

    |range| -- a tuple of two values, specifying the lower and
               the upper end of the interval spanned by the bins.
               Any data point outside this interval will be ignored.
               If no range is given, the smallest and largest
               data values are used to define the interval.

    The bin index and the number of points in a bin can be obtained by
    indexing the histogram with the bin number. Application of len()
    yields the number of bins. A histogram thus behaves like a
    sequence of (bin index, bin count) pairs.

    Here is an example on usage:
    
    >>> nsamples = 1000
    >>> import RandomArray
    >>> data = RandomArray.normal(1.0, 0.5, nsamples)
    >>> import Scientific.Statistics as S
    >>> S.mean(data)
    0.9607056871982641
    >>> S.standardDeviation(data)
    0.50251811830486681
    >>> S.median(data)
    0.94853870756924152
    >>> S.skewness(data)   # should be 0
    0.038940041870334556
    >>> S.kurtosis(data)   # should be 3
    2.865582791273765
    >>> 
    >>> from Scientific.Statistics.Histogram import Histogram
    >>> h = Histogram(data, 50)  # use 50 bins between min & max samples
    >>> h.normalizeArea()        # make probabilities in histogram
    >>> h[3]                     # bin index and frequency in the 4th bin
    array([-0.45791018,  0.01553658])
    >>> x = h.getBinIndices()
    >>> y = h.getBinCounts()
    >>> # can plot the y vector against the x vector (see below)
    >>> 
    >>> # add many more samples:
    >>> nsamples2 = nsamples*100
    >>> data = RandomArray.normal(1.0, 0.5, nsamples2)
    >>> h.addData(data)
    >>> h.normalizeArea()
    >>> x2 = h.getBinIndices()
    >>> y2 = h.getBinCounts()

    >>> import Gnuplot
    >>> g = Gnuplot.Gnuplot(persist=1)
    >>> g.xlabel('sample value');  g.ylabel('probability')
    >>> d1 = Gnuplot.Data(x, y, with='lines',
    ...                   title='%d samples' % nsamples)
    >>> d2 = Gnuplot.Data(x2, y2, with='lines',
    ...                   title='%d samples' % nsamples2)
    >>> g.plot(d1,d2)
    """
    
    def __init__(self, data, nbins, range=None):
        self._setup(data, nbins, range)
        self.addData(data)

    def _setup(self, data, nbins, range):
        if range is None:
            self.min = N.minimum.reduce(data)
            self.max = N.maximum.reduce(data)
        else:
            self.min, self.max = range
        self.min = self.min+0.
        self.max = self.max+0.
        self.bin_width = (self.max-self.min)/nbins
        self.array = N.zeros((nbins, 2), N.Float)
        self.array[:, 0] = self.min + self.bin_width*(N.arange(nbins)+0.5)

    def __len__(self):
        """Return the number of bins."""
        return self.array.shape[0]

    def __getitem__(self, index):
        """
        [i] returns (bin index, bin count) for bin no. i.
        [i,0] returns bin index for bin no. i.
        [i,1] returns bin count for bin no. i.
        Slices are also implemented.
        """
        return self.array[index]

    def __getslice__(self, first, last):
        return self.array[first:last]

    def getBinIndices(self):
        """
        Return an array of all the bin indices.
        Can conveniently be used to plot the histogram:
        >>> x = histogram.getBinIndices()
        >>> y = histogram.getBinCounts()
        Plot x against y.
        """
        return self.array[:,0].copy()
    
    def getBinCounts(self):
        """
        Return an array of all the bin counts.
        Can conveniently be used to plot the histogram:
        >>> x = histogram.getBinIndices()
        >>> y = histogram.getBinCounts()
        Plot x against y.
        """
        return self.array[:,1].copy()
        
    def addData(self, data):
        """Add the values in |data| (a sequence of numbers) to the
        originally supplied data. Note that this does not affect the
        default range of the histogram, which is fixed when the
        histogram is created.
        """
        n = (len(data)+999)/1000
        for i in range(n):
            self._addData(data[1000*i:1000*(i+1)])

    def _addData(self, data):
        data = N.array(data, N.Float)
        data = N.repeat(data, N.logical_and(N.less_equal(data, self.max),
                                            N.greater_equal(data, self.min)))
        data = N.floor((data - self.min)/self.bin_width).astype(N.Int)
        nbins = self.array.shape[0]
        histo = N.add.reduce(N.equal(N.arange(nbins)[:,N.NewAxis], data), -1)
        histo[-1] = histo[-1] + N.add.reduce(N.equal(nbins, data))
        self.array[:, 1] =  self.array[:, 1] + histo

    def normalize(self, norm=1.):
        "Scales all counts by the same factor such that their sum is |norm|."
        self.array[:, 1] = norm*self.array[:, 1]/N.add.reduce(self.array[:, 1])

    def normalizeArea(self, norm=1.):
        """Scales all counts by the same factor such that the area under
        the histogram is |norm|."""
        self.normalize(norm/self.bin_width)


class WeightedHistogram(Histogram):

    """Weighted histogram in one variable

    Constructor: WeightedHistogram(|data|, |weights|, |bins|, |range|=None)

    Arguments:

    |data| -- a sequence of data points

    |weights| -- a sequence of weights, same length as |data|

    |bins| -- the number of bins into which the data is to be sorted

    |range| -- a tuple of two values, specifying the lower and
               the upper end of the interval spanned by the bins.
               Any data point outside this interval will be ignored.
               If no range is given, the smallest and largest
               data values are used to define the interval.

    In a weighted histogram, each point has a specific weight. If all
    weights are one, the result is equivalent to a standard histogram.
    The bin index and the number of points in a bin can be obtained by
    indexing the histogram with the bin number. Application of len()
    yields the number of bins. A histogram thus behaves like a
    sequence of bin index - bin count pairs.
    """

    def __init__(self, data, weights, nbins, range=None):
        if len(data) != len(weights):
            raise ValueError, "wrong number of weights"
        self._setup(data, nbins, range)
        self.addData(data, weights)

    def addData(self, data, weights):
        """Add the values in |data| (a sequence of numbers) with the
        given |weights| to the originally supplied data. Note that
        this does not affect the default range of the histogram, which
        is fixed when the histogram is created.
        """
        if len(data) != len(weights):
            raise ValueError, "wrong number of weights"
        n = (len(data)+999)/1000
        for i in range(n):
            self._addData(data[1000*i:1000*(i+1)], weights[1000*i:1000*(i+1)])

    def _addData(self, data, weights):
        data = N.array(data, N.Float)
        weights = N.array(weights, N.Float)
        mask = N.logical_and(N.less_equal(data, self.max),
                             N.greater_equal(data, self.min))
        data = N.repeat(data, mask)
        weights = N.repeat(weights, mask)
        data = N.floor((data - self.min)/self.bin_width).astype(N.Int)
        nbins = self.array.shape[0]
        histo = N.add.reduce(weights*N.equal(N.arange(nbins)[:,N.NewAxis],
                                             data), -1)
        histo[-1] = histo[-1] + N.add.reduce(N.repeat(weights,
                                                      N.equal(nbins, data)))
        self.array[:, 1] =  self.array[:, 1] + histo


if __name__ == '__main__':
    nsamples = 1000
    import RandomArray
    RandomArray.seed(12,13)
    data = RandomArray.normal(1.0, 0.5, nsamples)
    h = Histogram(data, 50)  # use 50 bins between min & max samples
    h.normalizeArea()        # make probabilities in histogram
    x = h.getBinIndices()
    y = h.getBinCounts()

    # add many more samples:
    nsamples2 = nsamples*100
    data = RandomArray.normal(1.0, 0.5, nsamples2)
    h.addData(data)
    h.normalizeArea()
    x2 = h.getBinIndices()
    y2 = h.getBinCounts()

    # and more:
    nsamples3 = nsamples*1000
    data = RandomArray.normal(1.0, 0.5, nsamples3)
    h.addData(data)
    h.normalizeArea()
    x3 = h.getBinIndices()
    y3 = h.getBinCounts()

    import Gnuplot
    g = Gnuplot.Gnuplot(persist=1)
    g.xlabel('sample value');  g.ylabel('probability')
    d1 = Gnuplot.Data(x, y, with='lines',
                      title='%d samples' % nsamples)
    d2 = Gnuplot.Data(x2, y2, with='lines',
                      title='%d samples' % nsamples2)
    d3 = Gnuplot.Data(x3, y3, with='lines',
                      title='%d samples' % nsamples3)
    g.plot(d1, d2, d3)
