# -*- perl -*-
# Eps.pm $Id: Eps.pm,v 1.15.2.2 2000/01/30 13:56:23 jens Exp $
# (C) Copyright Jens G Jensen <jens@argaeus.ma.rhbnc.ac.uk>
# This file is part of epsmerge and is distributed under GNU GPL


package Eps;

use strict;
use Options;


# input: filename, IO ref, function to store procedures, function to
#	store DSCs
sub new {
    my ($class, $fname, $io, $store_proc, ) = @_;
    # An eps has boundingbox, orientation (default portrait) and a filename
    # - also knows name of save variable
    my $self = { fn => $fname, save => 'eps_Inc_Save_the_State' };
    bless $self, $class;

    unless( open(HELLO, "<$fname") ) {
	return "Creating $class: can't open $fname for reading\n";
    }
    $_ = <HELLO>;
    # File is expected to be either encapsulated postscript, or non-e postscript
    # in which case we pick *one* page (later) to display; see the ?? method below.
    # Note the distinction between `Pages:' (a DSC comment) and `pages' (a list of
    # page [numbers, labels, offset] refs in the order they should be displayed)
    if( /^%!PS-Adobe.*EPSF/i ) {
	$self->{'pages'} = undef;
    }
    elsif( /^%!/ ) {
	$self->{'pages'} = [ ];
    }
    else {
	return "It seems that $fname is not postscript\n";
    }

    # Simple parser: parse DSCs in the header looking in particular for the boundingbox.
    # Uh -- the parser is getting a bit ugly (and long)
    # atend says: 0 = ignore trailer, 1 = must read trailer, 2 = reading trailer,
    #	4 = reading procs
    my $atend = 0;
    # emblevel counts depth of embedded documents
    my $emblevel = 0;
    # Reading_List specifies which DSCs we are looking for and their parsed
    # values (keeping the first in the Header or the last in the Trailer).
    # Each entry has a name as the key, a regexp saying how to parse each entry of the
    # value, and sometimes a default value, and an expected length-of-list.
    my %Reading_List = (
		        'BoundingBox:' => { re => '\d+', ll => 4 },
		        'Orientation:' => { re => 'Portrait|Landscape', df => ['Portrait'], ll =>1 },
		        'LanguageLevel:' => { re => '\d+', df => ['3.0'], ll => 1 },
		        'Extensions:' => { re => '.+', },
			'Title:' => { re => '.*', df => ['(no title)'] },
			'CreationDate:' => { re => '.+', },
			'Pages:' => { re => '\d+', df => [1], ll => 1, },
			'PageOrder:' => { re => 'Ascend|Descend|Special', df => ['Ascend'], ll =>1 },
		       );
    # All but the last of these store locations in the file;
    # $name remembers the name of the DSC that we are currently reading.
    my ($end_of_header, $trailer,
	$begin_setup, $end_setup,
	$begin_prolog, $end_prolog,
	$name);
  PARSE:
    while( <HELLO> ) {
	--$emblevel if /^%%EndDocument/ || /^%%EndData/;
	++$emblevel if /^%%BeginDocument/ || /^%%BeginData/;
	next if $emblevel;
	# is this a %%+ continuation of the last DSC comment?
	if( /^%%\+\s(.+)$/ ) {
	    next if $name eq 'ignore';		# continuation of a comment we ignore
	    if( !defined $name || !defined $self->{$name} ) {
		print STDERR "Gakk! Found continuation, but I don't know of what!\n";
	    }
	    else {
		push @{$self->{ $name }}, split(/\s+/, $1);
	    }
	    next;
	}

	if( !defined $end_of_header && &_endofheader($_) ) {
	    $end_of_header = tell HELLO;
	}
	elsif(/^%%Trailer/) {
	    $trailer = tell HELLO;
	    $atend |= 2;
	    $atend &= ~1;
	    next;
	}
	elsif(/^%%BeginProlog/) {
	    $begin_prolog = tell HELLO;
	    $atend |= 4;
	    next;
	}
	elsif(/^%%EndProlog/) {
	    $end_prolog = tell HELLO;
	    $atend &= ~4;
	    next;
	}
	elsif(/^%%BeginProcSet/ || /^%%BeginResource/ || $atend & 4) {
	    # This stuff needed for sloppy files that don't bother to %%BeginProlog
	    $atend |= 4;
	    $begin_prolog = tell HELLO unless defined $begin_prolog;
	    # Get the procedure/resource
	    my ($what) = /^%%Begin(\w+)/;
	    my $proc = $_;
	    while(<HELLO>) {
		s/(^|\W)showpage(\W|$)/$1$2/;
		last if /^%%EndProlog/;
		$proc .= $_;
		last if defined $what && /^%%End$what/;
	    }
	    &$store_proc($proc);
	    $atend &= ~4 if /^%%EndProlog/;
	    next;
	}
	elsif(/^%%BeginSetup/) {
	    $begin_setup = tell HELLO;
	    next;
	}
	elsif(/^%%EndSetup/) {
	    $end_setup = tell HELLO;
	    next;
	}
	last if /^%%EOF/ && !Options->new()->getopts('ignore-eof');

	# identify comment
	my ($arg, $item);
	next unless ($name, $arg) = /^%%([a-zA-Z\.\?\!]+:?)\s*(.*)$/;
	# strip leading and trailing spaces
	$arg =~ s/^\s*//; $arg =~ s/\s*$//;
	$item = $Reading_List{ $name };
	next unless $item;
	if( $arg eq '(atend)' ) {
	    if( $atend & 2 ) {
		return "$fname: Huh? Can't have (atend) ref in the trailer!\n";
	    }
	    $atend |= 1;
	    $item->{ value } = '(atend)';
	    next;
	}
	if( $arg =~ /$item->{re}/ ) {
	    # The DSC that counts is the _first_ in _headers_, _last_ in _trailers_
	    # Here the trailer one (only if available and someone prompts us to read trailer)
	    # takes precedence over the header one
	    if( $self->{ $name } && !($atend & 2) ) {
		$name = 'ignore';
		next;
	    }
	    $self->{ $name } = [ split /\s+/, $arg ];
	}
	else {
	    return "$fname: Cannot read DSC $name";
	}
    }

    # Now we've done the header but we'll read on
    # to find all the pages if this is a non-E postscript.
    if( defined $self->{'pages'} ) {
	my $p = 0;
	if( ! defined $self->{'Pages:'} ) {
	    print STDERR "Warning: postscript file didn't tell me how many pages it had\n";
	}
	seek(HELLO, $end_of_header, 0) if defined $end_of_header;
	while(<HELLO>) {
	    if( /^%%Page: *(\S+) +(\d+)/ ) {
		push @{$self->{'pages'}}, [ $1, $2, tell HELLO ];
		++$p;
	    }
	}
	if( defined $self->{'Pages:'} && $self->{'Pages:'}->[0] != $p ) {
	    printf STDERR "Warning: $fname: expected %d page%s; found %d\n", \
		$self->{'Pages:'}->[0], $self->{'Pages:'}->[0] != 1 ? "s" : "", $p;
	}
	$self->{'Pages:'} = [ $p ];
    }

    close HELLO;
    return "$fname requires a trailer but does not have one" if $atend & 1;
    if( !defined $begin_prolog && defined $end_prolog ) {
	$begin_prolog = $end_of_header;
    }
    $self->{'filepos'} = [ $end_of_header, $trailer,
			   $begin_prolog, $end_prolog,
			   $begin_setup, $end_setup,
			 ];

    # Test data
    for (keys %Reading_List) {
	print STDERR "Warning: $fname: option $_ should have been specified in trailer\n"
	    if defined $self->{$_} && $self->{$_} eq '(atend)';
    }

    # check that everything from the reading-list was defined;
    # at least those that might be used as labels *must* be def'd
    foreach (keys %Reading_List) {
	my $ref = $Reading_List{$_};
	if( defined $self->{ $_ } ) {
	    # check that we got the expected number of entries
	    if( defined $Reading_List{ll} ) {
		if($Reading_List{ll} != @{$self->{ $_}}) {
		    printf STDERR "epsmerge: $fname: expected %d elements for %s found %d\n", \
			$Reading_List{ll}, $_, scalar @{$self->{ $_}};
		    if(defined $Reading_List{df}) {
			print STDERR "\t...since there is a default, I'd rather pick that\n";
			goto DEFAULT;
		    }
		}
	    }
	    next;
	}
      DEFAULT:
	# The thing was not defined, find a default
	if( defined $ref->{df} ) {
	    $self->{ $_ } = $ref->{df};
	}
	elsif( $_ eq 'CreationDate:' ) {
	    # get file's mtime (last modified)
	    my $date = (stat( $self->{fn} ))[9];
	    # if no mtime, shouldn't ever happen, use now()
	    $date = scalar localtime( $date ? $date : time() ) unless $date;
	    $self->{$_} = [ split /\s+/, $date ];
	}
    }


    if( defined $self->{'pages'} && $self->{'PageOrder:'}->[0] ne 'Special' ) {
	# sort pages on second entry (first is label, second is physical page number)
	my @list = sort { $a->[1] <=> $b->[1] } @{$self->{'pages'}};
	if( $self->{'PageOrder:'}->[0] eq 'Descend' ) {
	    $self->{'pages'} = [ reverse @list ];
	}
	else {
	    $self->{'pages'} = [ @list ];
	}
    }

    return $self;
}							# sub new

# This method returns undef if its file describes EPS; if it describes
# postscript the method returns a reference to the page-descriptor array
# which looks like this:
# [ [ 'i', 1, o1 ], [ 'ii', 2, o2 ], [ 'iii', 3, o3 ], [ '1', 4, o4 ]
# i.e., references to lists, and each entry has the label first and then
# the physical page number, and finally the offset in the file where
# that page may be found.  The list is sorted according to the
# %%PageOrder DSC.
# Oh, and please don't change any of the data; it is not protected
# (this is not C++, after all, is it?)

sub pages {
    my $self = shift;
    return $self->{'pages'};
}

# getlabel returns a label string as follows:
# f: filename
# d: the date of creation (er, of the eps file)
# T: the title of the eps file, as described by the DSC
sub getlabel {
    my $self = shift;
    $_ = shift;
    if( /f/ ) {
	return $self->{fn};
    }
    elsif( /d/ ) {
	return join( ' ', @{$self->{'CreationDate:'}} );
    }
    elsif( /T/ ) {
	return join( ' ', @{$self->{'Title:'}} );
    }
    else {
	print STDERR "$self->{fn}: Warning: unknown label specifier $_ ignored\n";
    }
    return;
}

sub box {
    my $self = shift;
    my @bbox = @{$self->{ 'BoundingBox:' }};
    my ($w, $h) = ($bbox[2]-$bbox[0], $bbox[3]-$bbox[1]);
    if( $self->{'Orientation:'}->[0] eq 'Landscape' ) {
	my $tmp = $w;
	$w = $h;
	$h = $tmp;
    }
    return ($bbox[0], $bbox[1], $bbox[0]+$w, $bbox[1]+$h);
}

# write must be passed (1) a boundingbox (integers) and
# (2) scaling parameters (possibly floats) and (3, optionally)
# a page number of the page it is supposed to print (for non-e postscript).
# It writes clipping things and the EPS file itself
sub write {
    my $self = shift;
    my ($llx, $lly, $urx, $ury, $xsc, $ysc) = @_;
    my $page = @_ > 6 ? $_[6] : undef;
    open(DATA, "<$$self{fn}") or return "Couldn't reopen $$self{fn}: $!";
    print <<HERE;
/$self->{save} save def
newpath $llx $lly moveto
$llx $ury lineto $urx $ury lineto $urx $lly lineto
closepath
clip
newpath
HERE
    # no support for 'PageOrientation' yet, but why would you have such a thing anyway?
    if( $self->{ 'Orientation:' }->[0] eq 'Landscape' ) {
	print "$llx $ury translate -90 rotate\n";
	my $tmp = $xsc; $xsc = $ysc; $ysc = $tmp;
    }
    else {
	print "$llx $lly translate\n";
    }
    print <<HERE;
$xsc $ysc scale
${$self->{'BoundingBox:'}}[0] neg ${$self->{'BoundingBox:'}}[1] neg translate
%Epsmerge included file: $$self{fn}
HERE
    ;

    # Probably put Setup stuff here
    _write_section(\*DATA, $self->{'filepos'}->[4], $self->{'filepos'}->[5]);

    if(defined $page) {
	# Locate data for the correct page; index is currently physical page no.
	# Support for physical page numbers that aren't the sequence 1..max
	my $npages = @{$self->{'pages'}};
	my $ref;
	if( $page >= $npages
	    || ($ref = $self->{'pages'}->[ $page-1 ])->[1] != $page )
		{
		    # ack! not in physical order or nonexistent page!
		    # but they are sorted (haha!), so we do a binary search
		    my ($lo, $hi) = (0, $npages);
		    while( $lo <= $hi ) {
			my $i = ($lo + $hi) >> 1;
			if( $self->{'pages'}->[$i]->[1] == $page ) {
			    # found the page anyway!
			    $ref = $self->{'pages'}->[$i];
			    goto FOUNDTHEPAGEANYWAY;
			}
			if( $self->{'pages'}->[$i]->[1] < $page ) {
			    $lo = $i+1;
			}
			else {
			    $hi = $i-1;
			}
		    }
		    return "epsmerge::Eps request for non-existent page $page in file $self->{fn}\n";
		  FOUNDTHEPAGEANYWAY:
		}
	print "%Including page $ref->[0] (physical $page)\n";
	seek(DATA, $ref->[2], 0);
    }
    else {
	# Including a std EPS file; skip the header
	seek DATA, $self->{'filepos'}->[0], 0;
    }
    my $atend = 0;
    my $emblevel = 0;
    while(<DATA>) {
	--$emblevel if /^%%EndDocument/;
	++$emblevel if /^%%BeginDocument/;
	next if $emblevel && /^%%/;		# ignore embedded documents' DSC
	if( /^%%Page/ ) {
	    if(defined $page) {
		last;				# done when we find next page
	    }
	    else {
		next;				# ignore embedded page thingies
	    }
	}
	# if we see `showpage' surrounded by whitespace, kill it and leave just the spaces
	# (doesn't kill `/showpage', nor does it do evil things to `myshowpageidentifier')
        s/(^|\W)showpage(\W|$)/$1$2/;
	if( /^%%Trailer/ ) {
	    $atend = 1;
	    if( defined $page ) {
		last;
	    }
	    else {
		next;
	    }
	}
	next if $atend && /^%%/;
	last if /^%%EOF/;
	print;
    }
    # Write the Trailer, hopefully cleaning up after the Setup code
    _write_section(\*DATA, $self->{'filepos'}->[1], -1);
    print "$self->{save} restore\n";
    close DATA;
    return;
}

# Return true if the input could be the first non-header line (*not* a method!)
sub _endofheader {
    my $data = shift;
    return ($data =~ /^%%EndComments/) || ($data !~ /^%\S/);
}

# Write a section of the file, given FILE *, start, end.  Ignore comments.
# end can be nonsense (like -1), in which case we read till EOF.
# However, if start or end are undefined, then nothing happens.
sub _write_section {
    my ($fh, $start, $end) = @_;
    return unless defined $start && defined $end;
    my $pos = tell $fh;
    seek $fh, $start, 0;
    until( eof($fh) || ($end != -1 && tell $fh > $end ) ) {
	$_ = <$fh>;
	next if /^%/;
	print;
    }
    seek $fh, $pos, 0;
}

package EpsWrap;

# package EpsWrap is a simple Eps-wrapper; it's interface is exactly the
# same as that of Eps.  It's raison d'etre is to have several EpsWrap'es
# sharing one Eps Cell such that if the Eps cell represents a postscript
# file then all EpsWrap's represent individual pages and they all share
# pointers to the same Eps.  This saves potentially lots of memory and
# time for the first-stage parsing of the file (only done once).

# input: eps cell reference and physical page number
sub new {
    my $class = shift;
    my $data = { 'eps' => $_[0], 'page' => $_[1] };
    return bless $data, $class;
}

sub pages {
    my $self = shift;
    return $self->{'eps'}->pages();
}

sub getlabel {
    my $self = shift;
    return $self->{'eps'}->getlabel(@_);
}

sub box {
    my $self = shift;
    return $self->{'eps'}->box(@_);
}

sub write {
    my $self = shift;
    return $self->{'eps'}->write(@_, $self->{'page'});
}

1;
