/*
 * $Id: cfb.c,v 1.3 2001/08/27 23:41:48 dave Exp $
 *
 * Copyright (C) 2001, David Pettersson 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *    * Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 *    * Neither the name of David Pettersson nor the names of his
 *      contributors may be used to endorse or promote products derived
 *      from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <Python.h>
#include <SDL/SDL.h>

/*
 * Definitions
 */

#define PACK(r, g, b) (0|(r)<<16|(g)<<8|(b))
#define BUFLEN 512

/*
 * Global
 */

static SDL_Surface *cfb;
static PyObject *cfbError;

/*
 * Forward
 */

int event_filter(const SDL_Event *event);

/*
 * Video stuff
 */

static void video_init(const int x, const int y, const char *caption)
{
  int flags = SDL_HWSURFACE|SDL_DOUBLEBUF;

  if(caption == NULL)
    flags |= SDL_FULLSCREEN;

  SDL_Init(SDL_INIT_VIDEO);
  SDL_SetVideoMode(x, y, 32, flags);
  SDL_SetEventFilter(event_filter);
  SDL_EnableUNICODE(1);
  
  if(caption)
    SDL_WM_SetCaption(caption, NULL);

  SDL_ShowCursor(SDL_ENABLE);
  cfb = SDL_GetVideoSurface();
}

static void video_putpixel(const int x, const int y, 
			   const int r, const int g, const int b)
{
  int col = PACK(r, g, b);
  int *fb = (int *) cfb->pixels;

  SDL_LockSurface(cfb);
  fb[y * cfb->w + x] = col;
  SDL_UnlockSurface(cfb);
}

static void video_clear(const int r, const int g, const int b)
{
  SDL_LockSurface(cfb);
  SDL_FillRect(cfb, NULL, PACK(r, g, b));
  SDL_UnlockSurface(cfb);
}

static void video_flip(void)
{
  SDL_Flip(cfb);
}

static void video_quit(void)
{
  SDL_Quit();
}

/*
 * Event
 */

int event_filter(const SDL_Event *event)
{
  switch(event->type)
    {
    case SDL_VIDEORESIZE:
    case SDL_SYSWMEVENT:
      video_flip();
      return 0;
      break;
    }

  return 1;
}

static PyObject *event_keyboard(SDL_KeyboardEvent *key)
{
  return Py_BuildValue("ss", "keyboard", SDL_GetKeyName(key->keysym.sym));
}

static PyObject *event_mouse(SDL_MouseButtonEvent *button)
{
  return Py_BuildValue("siii", "mouse", button->button, button->x, button->y);
}

static PyObject *event_quit(SDL_QuitEvent *quit)
{
  return Py_BuildValue("(s)", "quit");
}

/*
 * Python
 */

static PyObject *cfb_open(PyObject *self, PyObject *args)
{
  const char *caption = NULL;
  int x, y;

  if(!PyArg_ParseTuple(args, "ii|s", &x, &y, &caption))
    return NULL;

  video_init(x, y, caption);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *cfb_set(PyObject *self, PyObject *args)
{
  int x, y, r, g, b;
  
  if(!PyArg_ParseTuple(args, "iiiii", &x, &y, &r, &g, &b))
    return NULL;

  video_putpixel(x, y, r, g, b);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *cfb_clr(PyObject *self, PyObject *args)
{
  int r, g, b;

  if(!PyArg_ParseTuple(args, "iii", &r, &g, &b))
    return NULL;

  video_clear(r, g, b);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *cfb_flip(PyObject *self, PyObject *args)
{
  if(!PyArg_ParseTuple(args, ""))
    return NULL;

  video_flip();
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *cfb_close(PyObject *self, PyObject *args)
{
  if(!PyArg_ParseTuple(args, ""))
    return NULL;

  video_quit();
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *cfb_event(PyObject *self, PyObject *args)
{
  SDL_Event event;
  int done;

  if(!PyArg_ParseTuple(args, ""))
    return NULL;

  done = 0;

  while(done == 0)
    {
      SDL_WaitEvent(&event);

      switch(event.type)
	{
	case SDL_KEYDOWN:
	  return event_keyboard(&event.key);
	  break;
	case SDL_MOUSEBUTTONDOWN:
	  return event_mouse(&event.button);
	  break;
	case SDL_QUIT:
	  return event_quit(&event.quit);
	  break;
	}
    }

  Py_INCREF(Py_None);
  return Py_None;
}

static PyMethodDef cfbMethods[]= 
{
  { "open", cfb_open, METH_VARARGS },
  { "set", cfb_set, METH_VARARGS },
  { "clr", cfb_clr, METH_VARARGS },
  { "flip", cfb_flip, METH_VARARGS },
  { "close", cfb_close, METH_VARARGS },
  { "event", cfb_event, METH_VARARGS },
  { NULL, NULL }
};

void initcfb(void)
{
  PyObject *m, *d;
  
  m = Py_InitModule("cfb", cfbMethods);
  d = PyModule_GetDict(m);
  cfbError = PyErr_NewException("cfb.error", NULL, NULL);
  PyDict_SetItemString(d, "error", cfbError);
}

