#
# $Id: fb.py,v 1.4 2001/08/28 00:19:04 dave Exp $
#
# Copyright (C) 2001, David Pettersson 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#
#    * Neither the name of David Pettersson nor the names of his
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import cfb

class FrameBuffer:
    '''Class representing the frame buffer.

    This class gives access to a simple frame buffer. It is not a true
    frame buffer however, since you in most real ones have direct
    access to the video memory - the place where the pixels actually
    are stored.
    '''
    
    def __init__(self):
        self.fg = ( 255, 255, 255 )
        self.bg = ( 0, 0, 0 )
        self.size = (None, None)
        self.buffer = 1
        
    def open(self, x, y, caption, fullscreen = 0):
        '''Opens up an x by y window with an optional caption.

        If you want the window to run in full screen mode, simply pass
        a 1 as the last argument. This will make the window cover the
        entire screen. Beware thoguh, since it makes it difficult to
        exit the application on some computers (typically you cannot
        get back to your Python console.
        '''
        if fullscreen:
            cfb.open(x, y)
        else:
            if not caption: caption = ' '
            cfb.open(x, y, caption)

        self.size = ( x, y )

    def plot(self, x, y):
        '''Fills the pixel at (x, y).

        The color of the pixel is determined by the fg instance
        variable. Simply set the fg variable to a triple (red, green,
        blue).
        '''
        r, g, b = self.fg
        cfb.set(x, y, r, g, b)
        if not self.buffer: self.flip()
        
    def line(self, x1, y1, x2, y2):
        '''Draws a line from (x1, y1) to (x2, y2)..

        The color of the line is determined by the fg instance
        variable. Simply set the fg variable to a triple (red, green,
        blue).
        '''
        if y1 > y2:
            y1 ^= y2
            y2 ^= y1
            y1 ^= y2

            x1 ^= x2
            x2 ^= x1
            x1 ^= x2

        dx = x2 - x1
        dy = y2 - y1

        if dx > 0:
            dir = 1
        else:
            dx = -dx
            dir = -1

        if dx > dy:
            axis = 0
        else:
            axis = 1

        if axis == 0:
            ddy = dy + dy
            ddd = ddy - (dx + dx)
            err = ddy - dx

            self.plot(x1, y1)

            while dx > 0:
                if err >= 0:
                    y1 += 1
                    err += ddd
                else:
                    err += ddy
	  
                x1 += dir
                dx -= 1
                self.plot(x1, y1)
        else:
            ddx = dx + dx
            ddd = ddx - (dy + dy)
            err = ddx - dy

            self.plot(x1, y1)

            while dy > 0:
                if err >= 0:
                    x1 += dir
                    err += ddd
                else:
                    err += ddx

                y1 += 1
                dy -= 1
                self.plot(x1, y1)
    
    def clear(self):
        '''Clears the entire window and fills it with the background color.

        To set the background color, give the bg instance variable a
        triple containng the red, green and blue values).
        '''
        r, g, b = self.bg
        cfb.clr(r, g, b)

    def flip(self):
        '''Renders everything in the buffer so far.

        When you plot and draw lines, you do it in an off-screen
        buffer to improve performance. If the screen where to be
        updated after each pixel update the screen would flicker
        alot. Flipping the buffers should be done when you have
        finished drawing the current frame, or wish to see the results
        of the calls you have made so far. This is also known as
        double buffering and is a very common technique.

        It is possible to disable the buffer by setting the instance
        variable buffering to zero to disable the buffering. This is
        however not recommended.
        '''
        cfb.flip()
        
    def close(self):
        '''Closes the window opened by the open call.'''
        cfb.close()

    def event(self):
        '''Waits and returns a keyboard, mouse or window event.

        This is a method that will wait until it recieves an event
        from the operating system. An event is a keypress when the
        window is focused, as well as mouse clicks inside the window
        and on the close button.

        All events are returned as tuples, and the first element in
        the tuple is always a string describing what kind of event it
        is. Keyboard events return ('keyboard', key), where key is the
        key that was pressed. Mouse events return ('mouse', button, x,
        y), where button is what mouse button was pressed and x, y the
        x and y coordinates where the pointer was at the moment of the
        click. Also, there is a quit event that returns ('quit', ),
        and lets you know that the user tried to close the window.
        '''
        return cfb.event()
    
        
        
