#!/usr/local/bin/python

""" Module-Class-Importer for Python (Version 0.6)

    Modules in Python behave almost like classes,
    but do not provide the same mechanisms, like inheritance,
    baseclasses, special methods, etc.

    This module provides an alternative module loader, that
    is build on top of the ihooks.py-interface for the
    builtin import statement. It works in a similar way,
    the normal import does, but provides some extra features:

    * when a module is requested, an instance of the
      Module-class (or some subclass of Module) is created
      and the actions 'find' and 'load' are redirected to
      this instance via method calls
    * after loading, a call to install_module copies all the
      attributes from the "real" module object to the Module
      instance (which costs some memory, but increases lookup speed),
      thereby making it behave just like the original
    * a reference to the original module object is kept,
      so that 'from...import...' also works (since this statement
      needs a real module object)
    * whenever a module is referenced, the Module object is
      returned, if possible, so even after having done 'from x import y'
      at some point, 'import x' will return the Module object,
      so hopefully all references to a module made by a Python
      program should return the Module object, with all its
      nice advantages (like catching AttributeErrors)
    * Module provides a basic skeleton -- you can subclass
      it and then give the ModuleClassImporter your class to use
      (LazyModule is an example for this), if you don't like
      some things, like copying attributes (e.g. use __getattr__
      to redirect the lookup)

    This module contains all necessary base classes (working ones,
    not simply a bare framework), some Loaders, and
    of course, the LazyModule which started this whole thing in
    the first place.

    For more information on how importing works, see ihooks.py
    and ni.py.

    ----------------------------------------------------------------

    Example of usage: Lazy Import for Python (see LazyImp.py)

    ---------------------------------------------------------------

    History:
    - 0.6: fixed for Python 1.5

    Bugs: 
    - none, only unsupported features :-)
    - I have tested it with Tkinter and a 10.000 line framework,
      but of course... there may still be some imports out there,
      I haven't taken into account yet.

    
-----------------------------------------------------------------------------
(c) Copyright by Marc-Andre Lemburg (mailto:mal@lemburg.com)

    Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee or royalty is hereby granted,
    provided that the above copyright notice appear in all copies and that
    both that copyright notice and this permission notice appear in
    supporting documentation or portions thereof, including modifications,
    that you make.

    THE AUTHOR MARC-ANDRE LEMBURG DISCLAIMS ALL WARRANTIES WITH REGARD TO
    THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
    FITNESS, IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
    INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
    FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
    NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
    WITH THE USE OR PERFORMANCE OF THIS SOFTWARE !


"""

__version__ = '0.6'

import sys,ihooks,imp,os

# so that it also works under Python 1.4
try:
    __debug__
except:
    __debug__ = 0

#
# A fast ModuleLoader
#

class FastModuleLoader(ihooks.ModuleLoader):

    """ works like ModuleLoader, but uses imp's find_module, which makes
        it somewhat faster
        
        * note: file system hooks won't work here !!!
    """
    def find_module(self,name,path=None): 
        m = self.find_builtin_module(name)
        if m: return m
        if path is None: path = sys.path
        return imp.find_module(name,path)

#
# A preprocessing loader
#
# (parts taken from py_compile.py)

import marshal

def clong(x):
    """ return the 4-byte long x as 4-byte string """
    return chr(x&0xff)+chr((x>>8)&0xff)+chr((x>>16)&0xff)+chr((x>>24)&0xff)

class PreProcessingLoader(FastModuleLoader):

    """ do some preprocessing when importing a module, that
        has to be compiled first, i.e. is read in as source file
        * leaves the rest to FastModuleLoader 
    """

    def load_module(self, name, stuff):

        """ load the module name using stuff """

        file, filename, (suff, mode, type) = stuff
        # check if there already is a properly compiled version
        pass
        # if we have to handle a source file...
        if type == imp.PY_SOURCE:
            # read file
            program = file.read()
            # process program
            program = self.preprocess(program)
            # compile and try to write the .pyc-file (copied from py_compile.py)
            code = compile(program, filename, 'exec')
            codefilename = filename + (__debug__ and 'c' or 'o')
            try:
                fc = open(codefilename,'wb')
                fc.write(imp.get_magic())
                timestamp = long(os.stat(filename)[8])
                fc.write(clong(timestamp))
                marshal.dump(code,fc)
                fc.close()
                if os.name == 'mac':
                    import macfs
                    macfs.FSSpec(codefilename).SetCreatorType('Pyth', 'PYC ')
                    macfs.FSSpec(filename).SetCreatorType('Pyth', 'TEXT')
            except IOError:
                pass
        else:
            return FastModuleLoader.load_module(self, name, stuff)
        # register and initialize module
        m = self.hooks.add_module(name)
        m.__file__ = filename
        exec code in m.__dict__
        return m

    def preprocess(self,program):

        """ do something with the code in program and return
            the modified string
        """
        program = "The_PreProcessingLoader_was_here = ':-)'\n" + program
        return program

#
# The Module base class
#

class InternalVars: # container class
    pass

class Module:

    """ The module-works-as-a-class base class

        * this class is instantiated for every new module loaded
          by the SimulateImport mechanism
        * you can subclass the class to add functionality and
          pass the subclass to SimulateImport for it to be used
        * important: local variables should always reside in
          self.__moduleobj__, not in self directly (to avoid name
          clashes)
        * note: module initialization is done in the usual way, the
          modules namespaces then copied to this object
        * this class emulates the normal import-operation  
    """
    def __init__(self,name,loader,fromlist=None):

        """ a module name is requested 

            * this method should NOT be overridden, instead override
              startup() which is called, when this method finishes
        """
        self.__moduleobj__ = m = InternalVars()
        self.__name__ = name
        m.loader = loader
        m.fromlist = fromlist
        m.found = 0
        m.loaded = 0
        m.modules = loader.modules_dict()
        m.self = self
        m.module = None # gets filled by load_module()
        self.startup()

    def startup(self):

        """ module startup

            * called when a module is requested
        """
        self.find_module()
        self.load_module()

    def real_module(self):

        """ return a real module object """

        return self.__moduleobj__.module

    def register(self):

        """ makes an entry in modules pointing to this object 
            * loading a module through the loader normally also
              registers the module, so a call to this method is
              not needed
            * note: if you want to do 'from..import..' with
              this module later on, the registering MUST be done
              by loader
        """
        self.__moduleobj__.modules[self.__name__] = self

    def find_module(self):

        """ find the module """
        
        m = self.__moduleobj__
        m.stuff = m.loader.find_module(self.__name__)
        if not m.stuff: 
            raise ImportError, 'Module: No module named %s'%name
        m.found = 1

    def load_module(self):

        """ load the module and initialize it

            * the module must already be found
            * uses __moduleobj__.loader for loading
            * calls .install_module to complete the job
        """

        m = self.__moduleobj__
        if m.loaded: return
        if not m.found:
            raise ImportError, 'Module: call %s.find_module() first'%self.__name__
        else:
            module = m.loader.load_module(self.__name__,m.stuff)
        self.install_module(module)
        m.loaded = 1

    def install_module(self,module):

        """ install the module in this objects namespace 

            * must be called after a module is loaded
        """

        # keep a reference to the original
        self.__moduleobj__.module = module
        # copy all module attributes to this object
        for k,v in module.__dict__.items():
            setattr(self,k,v)
        # create a reference in the real module object
        setattr(module,'__moduleobj__',self.__moduleobj__)
        
    def __repr__(self):

        """ return some meaningful string describing self """

        if self.__moduleobj__.loaded: 
            return "<%s '%s'>"%(self.__class__.__name__,self.__name__)
        elif self.__moduleobj__.found:
            return "<%s '%s', loading deferred>"%(self.__class__.__name__,self.__name__)
        else:
            return "<%s '%s', finding deferred>"%(self.__class__.__name__,self.__name__)

    __str__ = __repr__

    def __getattr__(self,x):

        """ some unknown attribute is being requested """

        raise AttributeError,'%s "%s" was looking for "%s"'%(self.__class__.__name__,self.__name__,x)

#
# The module-as-class importer
#

# ModuleImporter to be used:
ImporterBaseClass = ihooks.ModuleImporter

class ModuleClassImporter(ImporterBaseClass):

    """ Module importer, that knows how to handle Module-objects correctly
    """

    def __init__(self,module_class,*importer_class_init):

        """ import modules by encapsulating them in an instance of
            module_class
            * modules_class must be a subclass of Module
            * the other parameters are passed to the ImportClass
              (see ihooks.py for details)
        """    
        apply(ImporterBaseClass.__init__,(self,)+importer_class_init)
        self.module_class = module_class

    def import_module(self, name, globals={}, locals={}, fromlist=None):

        """ module import hook
        """
        if self.modules.has_key(name): # fast path
            m = self.modules[name]
            # return the object, if possible
            if fromlist is None:
                #print 'Importer: import',name,'(found in sys.modules)',m
                try:
                    return m.__moduleobj__.self
                except:
                    return m
            else:
                # from..import.. insists on having the real thing !
                #print 'Importer: from',name,'import',fromlist,'(found in sys.modules)',m
                try:
                    return m.__moduleobj__.module
                except:
                    return m
        else:
            if fromlist is None:
                # normal 'import modulename'
                #print 'Importer: import "%s" with %s'%(name,self.module_class.__name__)
                module = apply(self.module_class,(name,self.loader,fromlist))
            else:
                # emulate 'from modulename import something'
                # (note: this a hack... and not a nice one !)
                #print 'Importer: from',name,'import',fromlist,'with',self.module_class.__name__
                module = apply(self.module_class,(name,self.loader,fromlist))
                # module has to be loaded for this to work
                module.load_module()
                module = module.real_module()
            #print 'Importer: %s returned %s'%(self.module_class.__name__,module)
            return module

