#!/usr/local/bin/python

"""
    Lazy Import for Python (Version 0.6)

    Loads modules only if they are needed and referenced.
    This is done by overloading the builtin 'import'
    statement, so no code changes are necessary.
    Everything should work as normal, except that the
    actual loading process is deferred until a module's
    attribute is requested (you have to keep in mind,
    that this can cause exceptions from the module 
    initialization process -- use lazyimp after debugging !)

    * depends on the module ClassModules.py

    *** Importing this module autoinstalls the Lazy Import Feature.
    *** All subsequent imports will be done lazy.
    *** If you don't like this, comment out the last line !

    For more information, see the LazyModule-doc string below.

    ---------------------------------------------------------------

    History:
    - 0.6: fixed for Python 1.5

    Bugs: 
    - none, only unsupported features :-)
    - I have tested it with Tkinter and a 10.000 line framework,
      but of course... there may still be some imports out there,
      I haven't taken into account yet.

    
-----------------------------------------------------------------------------
(c) Copyright by Marc-Andre Lemburg (mailto:mal@lemburg.com)

    Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee or royalty is hereby granted,
    provided that the above copyright notice appear in all copies and that
    both that copyright notice and this permission notice appear in
    supporting documentation or portions thereof, including modifications,
    that you make.

    THE AUTHOR MARC-ANDRE LEMBURG DISCLAIMS ALL WARRANTIES WITH REGARD TO
    THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
    FITNESS, IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
    INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
    FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
    NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
    WITH THE USE OR PERFORMANCE OF THIS SOFTWARE !


"""

__version__ = '0.6'

import sys,ihooks,imp,os
from ClassModules import *

# base class to be used:
LazyModuleBaseClass = Module

class LazyModule(LazyModuleBaseClass):

    """ Lazy Import for Python

    Loads modules only if they are needed and referenced.
    This is done by overloading the builtin 'import'
    statement, so no code changes are necessary.
    Everything should work as normal, except that the
    actual loading process is deferred until a module's
    attribute is requested (you have to keep in mind,
    that this can cause exceptions from the module 
    initialization process -- use lazyimp after debugging !)

    Hints:
    - you can call the method load_module() of a lazy module
      to force loading of the module (or simply reference
      some attribute)

    Caveats:
    - attributes like __dict__ and __name__, that are provided
      by the LazyImport-class, do not cause loading
    - due to a Python internal limitation, from ... import ...
      is not handled in a lazy fashion (wouldn't be too efficient anyway)
    - debugging circular imports can become an even harder task
      (uncomment the #print-statements to see what's going on)
    """  

    # finding the module is normally done when the object is created
    # -- setting this to 1 defers finding too
    __defer_find = 0

    def startup(self):

        """ lazy import module
        """
        self.__moduleobj__.defer_find = self.__defer_find
        if not self.__moduleobj__.defer_find: 
            self.find_module()
        self.register()

    def load_module(self,cause='*'):

        """ do the actual import

            * this can cause ImportErrors and raise exceptions, that
              must be handled by the caller, i.e. the first reference
              to a module might raise an exception !
            * modules are only loaded once; any subsequent calls to this method
              are silently ignored (i.e. ImportErrors are only raised
              the first time, this method is used)
        """
        if self.__moduleobj__.loaded: return
        #print 'LazyModule: loading module "%s", looking for "%s" ...'%(self.__name__,cause)
        if self.__moduleobj__.defer_find: 
            # find now
            self.find_module()
        # let the base class handle the rest
        LazyModuleBaseClass.load_module(self)

        #print 'LazyModule: module "%s" loaded'%self.__name__

    def __getattr__(self,x):

        """ the module's needed, so load it and return the
            requested attribute afterwards
        """
        #print self.__name__,'is looking for',x
        if not self.__moduleobj__.loaded:
            self.load_module(x)
            return getattr(self,x)
        else:
            raise AttributeError,'%s "%s" was looking for "%s"'%(self.__class__.__name__,self.__name__,x)

def autoinstall():

    """ install the Lazy Module Import feature """

    mloader = FastModuleLoader()
    mhandler = LazyModule
    newimport = ModuleClassImporter(mhandler,mloader)
    newimport.install()

#
# auto-install as new 'import' (comment out, if you don't like this)
#
autoinstall()
