#!/usr/bin/env python
#----------------------------------------------------------
#
# Introduction :
#
# PythonPlot is a Python application for plotting x-y-graphs read from ASCII files,
# which have numbers, sorted in columns but may also have various,
# unformatted headers and comments. It's main purpose is to quickly analyse
# complicated curves by viewing them and zooming into interesting regions of
# the curves. For zooming in, you just drag a rectangle over the region of interest.
# PythonPlot opens text files first in a separate window where you can see the whole text,
# except rows and columns of numbers (they are truncated after the first few lines).
# In this text window, you can specify the start-line, the column for the x-axis and the
# column for the y-axis by two clicks with the left mouse button. Just click over the
# first value of the x-axis and then over the first value of the y-axis. PythonPlot then
# reads the numbers from the text file, until it finds a blank line or a non-number-character.
# Finally you can add some annotations to the curves and export the plot canvas to a
# postscript file. 
# PythonPlot is Open-Source-Freeware and all improvements and suggestions are welcome.
# Just look for it at the Vaults of Parnassus (http://www.vex.net/parnassus/)
# 
# Release notes :
# 
# Version 0.1.5, September 24, 2002
#   first released version by Bernd Aberle
#   Features : see introduction above
#   Known problems :
#     - no horizontal annotation text possible, I've read somewhere that it's still
#       missing in the canvas widget of Tkinter.
#
# Version 0.1.6, September 27, 2002
#   more comments added and help text introduced
#
# Version 0.1.7, November 19, 2002
#   x-y-axis labels improved
#
# Version 0.1.8, November 21, 2002
#   cancelling annotation adding caused an exception
#   recent file list reading/saving improved
#
# Version 0.1.9, November 27, 2002
#   using filename as title in ASCII-file-windows
#
# Version 0.2.0, January 26, 2003, Bernd Aberle
#   plotFile method opens now an ASCII-file-window
#
# Version 0.2.1, January 27, 2003
#   catched : AsciiDataFile instance has no attribute 'xminplot' if data not yet
#   loaded (data will be loaded when user clicks on plot)
#
# Version 0.2.2, April 26, 2003
#   The default plot range is now defined by the total minimum and maximum of
#   all curves. Methods getMinMaxOfAllCurves and getPlotRegion added
#   Tick-labels have now two digits
#   Default directory for UNIX (posix) in file-open-box is now current directory
#
# Version 0.2.3, Mai 8, 2003
#    "New" Button removed from menu (it was obsolete and confusing)
#    dialog = AnnotationDialog(None) replaced by dialog = AnnotationDialog(self.master)
#    (didn't work in OpenWindows)
#    Annotation text made persistent between calls of AnnotationDialog
#----------------------------------------------------------

VERSION = '0.2.3'

HELP_TEXT = \
  'Usage :\n \
  If you have Python and Tkinter, either double-click on PythonPlot.py in your file-browser or\n \
  start it from the command line:\n\n \
  python PythonPlot.py\n \
  or with parameters :\n \
  python PythonPlot.py filename startline xcolumn ycolumn\n\n \
  The default parameters are first line, first column for x-axis and first column for y-axis\n \
  (python PythonPlot.py filename 0 0 0)\n \
  If the columns for x- and y-axis are identical, PythonPot plots the graph over its index.\n \
  \n \
  The graphical user-interface is as simple as possible : open the ASCII file (File->Open),\n \
  select the x-axis by clicking on the first number of the x-axis, select the y-axis the same way\n \
  and click on the plot button.\n \
  You may open several ASCII files this way for overplotting.\n \
  To zoom in, drag a rectangle in the plot window, to zoom out a small amount, open the context\n \
  menu (right button click) and click on <zoom out x-axis> or <zoom out y-axis>, to zoom out\n \
  completely, double-click with the left mouse button.\n \
  Use the context menu for inserting or deleting annotations and\n \
  save the view as postscript via File->Save as postscript. Thats it !'

from Tkinter import *
import tkMessageBox
import tkFileDialog
import tkFont
from ScrolledText import ScrolledText
from math import *
import os
import os.path
import fpformat
import tkSimpleDialog

#----------------------------------------------------------
# Configuration class, holds all configuration parameters and
# constants (e.g. version number). Some persistant settings
# shall be saved at program exit and loaded at program start,
# e.g. recently opened files, etc
#----------------------------------------------------------
class MyConfig:

  #----------------------------------------------------------
  # read persistant options (e.g. recent files)
  #----------------------------------------------------------
  def readIniFile(self):

    try:
      file = open(self.inifilename, "r")
    except:
      return

    line = "anything"
    irecent = 0
    iline = 0
    while line != "":
      # first line is default directory for file-open-box
      line = file.readline()
      line = line.rstrip()
      if iline == 0 and line.strip() != "":
        self.filepath = line
        if os.name == 'posix':
          # in UNIX we prefer opening files always in current directory
          # because PythonPlot is mostly started from the command line
          # in the directory where the ASCII files to be plotted are
          self.filepath = None
      # now follow recently opended files (complete path)
      if iline >= 1:
        if irecent < self.numberOfRecentFiles and line.strip() != "":
          self.addRecentFile(line, irecent)
          irecent = irecent + 1
      iline = iline + 1
    file.close()

  #----------------------------------------------------------
  # write persistant options (e.g. recent files)
  #----------------------------------------------------------
  def writeIniFile(self):

    if self.inifilename != '':
      try:
        file = open(self.inifilename, "w")
        file.write(self.filepath + '\n')
        for irecent in range(self.numberOfRecentFiles) :
          file.write(self.recentfiles[irecent] + '\n')
        file.close()
      except:
        pass

  #----------------------------------------------------------
  # add a file to the recent files list, but only if it's not
  # already in the list
  #----------------------------------------------------------
  def addRecentFile(self, filename, index = 0):
    try:
      # if it is not already there
      self.recentfiles.index(filename)
    except:
      # insert it as first element and remove the last element
      self.recentfiles.insert(index, filename)
      self.recentfiles.pop()

  #----------------------------------------------------------
  # Constructor
  #----------------------------------------------------------
  def __init__(self):

    self.name = 'PythonPlot'
    self.version = VERSION

    # set the directory for opening files self.filepath
    # initially use the prefix of PythonPlot.py
    # self.filepath = sys.prefix # this is the prefix of Python
    pythonPlotPrefix = sys.argv[0]
    pythonPlotPrefix = os.path.dirname(pythonPlotPrefix)
    self.filepath = pythonPlotPrefix
    if os.name == 'posix':
      # in UNIX we prefer opening files in current directory
      self.filepath = None

    self.numberOfRecentFiles = 8
    self.recentfiles = [' '] * self.numberOfRecentFiles

    # if there is a home directory, use it for the initialisation file
    # else use the prefix of PythonPlot.py
    self.inifilename = ''
    self.home = os.environ.get('HOME')
    if self.home and os.path.isdir(self.home):
      self.inifilename = os.path.normpath(self.home + "/.PythonPlot.ini")
    else:
      if pythonPlotPrefix and os.path.isdir(pythonPlotPrefix):
        self.inifilename = os.path.normpath(pythonPlotPrefix + "/PythonPlot.ini")


#----------------------------------------------------------
# structure for annotations
#----------------------------------------------------------

# global for remembering the last entries between annotations
INITIAL_ANNO_TEXT = " "
INITIAL_ANNO_SIZE = "12"

class Annotation:
  #----------------------------------------------------------
  def __init__(self):
    self.x = 0
    self.y = 0
    self.mybbox = None
    self.textwidget = None
    self.color = "black"
    self.orientation = "horizontal"
    self.charsize = "12"
    self.text = ""


#----------------------------------------------------------
# Dialog for defining annotations, opens a window where
# you can specify a text, it's color and size. After closing
# the dialog you can position the annotations with the next
# mouse left-click
#----------------------------------------------------------
class AnnotationDialog(tkSimpleDialog.Dialog):

  # any canvas event
  # draw the annotation text in the preview canvas
  def annoCanvasCallback(self, event):
    try:
      self.currentcolor = self.plotcolor[int(event.y / 12)]
    except:
      pass
    self.annoButtonCallback()
    
  # any button event
  # draw the annotation text in the preview canvas
  def annoButtonCallback(self):
    if self.e1.get() != "" and self.e2.get() != "":
      self.preview.create_rectangle(0, 0, 99, 29, fill="white", outline="white")
      self.preview.create_text(10, 15, text=self.e1.get(), font=tkFont.Font(size=self.e2.get()),
        fill=self.currentcolor, anchor=W)
    
  # define the widgets
  def body(self, parent):

    self.l1 = Label(parent, text='text:')
    self.e1 = Entry(parent, width=60, text=INITIAL_ANNO_TEXT)
    
    self.e2var = StringVar()
    self.e2var.set(INITIAL_ANNO_SIZE)
    self.l2 = Label(parent, text='size:')
    self.e2 = Entry(parent, width=4, text=INITIAL_ANNO_SIZE, relief=SUNKEN, textvariable=self.e2var)
    
    self.l9 = Label(parent, text='to place the text use left mouse\nbutton after pressing "ok"')
    self.canvas = Canvas(parent, width=100, height=7*12, bg="white")
    self.b1 = Button(parent, text='Preview:', command=self.annoButtonCallback)
    self.preview = Canvas(parent, width=100, height=30, bg="white")

    # event bindings
    self.canvas.bind("<Button-1>", self.annoCanvasCallback)

    # avoid leaving the dialog if <return>was pressed
    # and allow leaving only via <ok> or <cancel> button
    self.e1.bind("<Return>", lambda e: "break")
    self.e2.bind("<Return>", lambda e: "break")

    self.plotcolor = ["black", "red", "blue", "magenta", "darkblue", "darkgreen", "brown"]
    for idx in range(len(self.plotcolor)):
      self.canvas.create_text(6, idx*12+6, text=self.plotcolor[idx], fill=self.plotcolor[idx], anchor=W)
    self.currentcolor = "black"
      
    self.l1.pack()
    self.e1.pack()
    self.l2.pack()
    self.e2.pack()
    self.canvas.pack()
    self.b1.pack()
    self.preview.pack()
    self.l9.pack()

    return self.e1 # initial focus

  # set the annotation structure data
  def apply(self):
    self.anno = Annotation()
    self.anno.color = self.currentcolor    
    if self.e2.get() != "":
      self.anno.charsize = self.e2.get()
      INITIAL_ANNO_SIZE = self.e2.get()
    self.anno.text = self.e1.get()
    INITIAL_ANNO_TEXT = self.anno.text
    self.result = self.anno

#----------------------------------------------------------
# This class reads and holds data from an ASCII file
# It needs start-line, column index for x-axis and column
# index for y-axis
#----------------------------------------------------------
class AsciiDataFile:

  #----------------------------------------------------------
  # jump to start-line and read all lines until a non-number
  # appears. From the read lines store column defined be
  # ixcolumn as x-value and the column defined by iycolumn
  # as y-value
  #----------------------------------------------------------
  def readData(self, startline = 0, ixcolumn = 0, iycolumn = 0):

    if self.filename.strip() == '':
      return

    self.xaxis = []
    self.yaxis = []

    readit = 0
    iline = 0 # line index of the file
    index = 0 # index of the copied data
    file = open(self.filename, "r")
    line = "anything"
    
    while line != "":
      line = file.readline()

      if iline == startline: # search given start line
        readit = 1

      if readit == 1: # search first line with numbers
        try:
          # print line.rstrip()
          parts = line.split()
          for ftmp in parts[:]:
            # throws exception if non-number
            ftmp1 = float(ftmp)
          x, y = float(parts[ixcolumn]), float(parts[iycolumn])
          readit = 2 # copying successful, so start to copy from now on
        except:
          pass
        
      if readit == 2: # copy these data
        try:
          parts = line.split()
          for ftmp in parts[:]:
            # throws exception if non-number
            ftmp1 = float(ftmp)
          x, y = float(parts[ixcolumn]), float(parts[iycolumn])
          if ixcolumn == iycolumn:
            # if columns are identical, use index as x-axis
            x = float(index)
            index = index + 1
          self.xaxis.append(x)
          self.yaxis.append(y)
        except:
          readit = 0

      iline = iline + 1
      
    file.close()

    # the minima and maxima of the curve
    self.xmin = min(self.xaxis)
    self.xmax = max(self.xaxis)
    self.ymin = min(self.yaxis)
    self.ymax = max(self.yaxis)

    # the minima and maxima of the part of the curve to be plotted
    self.xminplot = min(self.xaxis)
    self.xmaxplot = max(self.xaxis)
    self.yminplot = min(self.yaxis)
    self.ymaxplot = max(self.yaxis)

    if self.xmin == self.xmax:
      self.xmax = self.xmin + 1.0
    if self.ymin == self.ymax:
      self.ymax = self.ymin + 1.0

  #----------------------------------------------------------
  def __init__(self, filename):
    self.filename = filename
    self.xaxis = []
    self.yaxis = []


#----------------------------------------------------------
# Here we open the text window for an ASCII file and wait for
# the user to specify start-line, x-column and y-column. With
# these informations we set up an AsciiDataFile object
#----------------------------------------------------------
class AsciiDataFileAnalyser(Toplevel):

  #----------------------------------------------------------
  # if the window closes, destroy the data
  #----------------------------------------------------------
  def onExit(self):
    # if tkMessageBox.askokcancel("Quit", "Do you really wish to quit?"):
    try:
      self.plotwindow.filedata.remove(self.asciifile)
    except:
      pass
    self.destroy()

  #----------------------------------------------------------
  def plot(self):
    print self.startline, self.ixcolumn, self.iycolumn
    self.asciifile.readData(self.startline, self.ixcolumn, self.iycolumn)
    self.plotwindow.plot()

  #----------------------------------------------------------
  # determine the word columns for x and y values from the
  # clicked character columns
  #
  def findXYcolumns(self):
    if self.filename.strip() == '':
      return

    file = open(self.filename, "r")
    line = "anything"
    readit = 0
    iline = 0
    while line != "":
      line = file.readline()

      if iline == self.startline:
        # print "starting at:", line.rstrip()
        readit = 1
        subline = line[0:self.column_index + 1]
        subparts = subline.split()
        if self.toggle_xy == 'x':
          self.ixcolumn = len(subparts) - 1
        else:
          self.iycolumn = len(subparts) - 1
        print 'startline', self.startline, 'xcolumn', self.ixcolumn, 'ycolumn', self.iycolumn

      if readit == 1:
        try:
          parts = line.split()
          x, y = float(parts[self.ixcolumn]), float(parts[self.iycolumn])

          if iline < self.startline + 2:
            print x, y
          if iline == self.startline + 2:
            print '...'
            break # the column indizes should be set now

        except:
          readit = 0

      iline = iline + 1

    if self.toggle_xy == 'x':
      self.toggle_xy = 'y'
    else:
      self.toggle_xy = 'x'

    file.close()

  #----------------------------------------------------------
  # convert pixel coordinates into line/column index
  #
  def onLeftClick(self, event):
    
    line_column = self.text.index("@%d,%d" % (event.x, event.y))
    line_column = line_column.split('.')
    lineindex = int(line_column[0])
    if len(self.lineoffsets) == 0:
      self.startline = max(0, lineindex - 1)
    else:
      self.startline = max(0, self.lineoffsets[lineindex] - 1)
    self.column_index = int(line_column[1])

    # print 'startline, column_index', self.startline, self.column_index
    self.findXYcolumns()
    
  #----------------------------------------------------------
  # read the ASCII file and show it's content in the text widget
  # Don't show the complete number columns, only the first 5 lines
  # otherwise we would read to much unrelevant data if it's a big file
  # (we are no text-editor)
  #----------------------------------------------------------
  def readFile(self):

    if self.filename.strip() == '':
      return

    self.lineoffsets = []
    iline = 0 # line index of the file
    number_counter = 0
    max_number_lines = 5
    file = open(self.filename, "r")
    file.seek(0, 2) # move to file end
    filesize = file.tell()
    file.seek(0) # move to file begin

    if filesize < 10000:
      # for small files, insert all lines and don't use lineoffsets
      self.lineoffsets = []
      self.text.insert('end', file.read())

    else:
      # for big files, insert only text and few number lines and remember lineoffsets
      line = "anything"
      
      while line != "":
        line = file.readline()

        try:
          parts = line.split()
          for ftmp in parts[:]:
            # throws exception if non-number
            ftmp1 = float(ftmp)
          if number_counter < max_number_lines:
            self.text.insert('end', line)
            self.lineoffsets.append(iline)
          if number_counter == max_number_lines:
            self.text.insert('end', '...\n')        
            self.lineoffsets.append(self.lineoffsets[-1])
          number_counter = number_counter + 1
        except:
          self.text.insert('end', line)
          self.lineoffsets.append(iline)
          number_counter = 0
          
        iline = iline + 1
        
    file.close()

  #----------------------------------------------------------
  # Initialise the window with a scrolled text widget
  #----------------------------------------------------------
  def __init__(self, filename, plotwindow, asciifile):

    Toplevel.__init__(self)

    self.filename = filename
    self.plotwindow = plotwindow
    self.asciifile = asciifile

    self.ixcolumn = 0
    self.iycolumn = 0
    self.startline = 0
    self.toggle_xy = 'x'
    parts = filename.split('/')
    self.title(parts[len(parts) - 1])
    
    #----------------------------------------------------------
    # Widgets
    #----------------------------------------------------------
    font = tkFont.Font(size=12)
    # self.text = ScrolledText(self, bg="white", font=font, wrap='none', width=80, height=24)
    self.text = ScrolledText(self, bg="white", font=font)

    # horizontal scrollbar - this obviously shrinks the window ?!
##    self.hbar = Scrollbar(self.text, orient='horizontal')
##    self.text['xscrollcommand'] = self.hbar.set
##    self.hbar['command'] = self.text.xview
##    self.hbar.pack(side=BOTTOM, fill=X)

    self.bplot = Button(self, text="Plot",  fg="blue", command=self.plot)
    
    #----------------------------------------------------------
    # Event bindings
    #----------------------------------------------------------
    self.text.bind("<Button-1>", self.onLeftClick)
    self.protocol("WM_DELETE_WINDOW", self.onExit)
    
    #----------------------------------------------------------
    # Layout
    #----------------------------------------------------------
    self.bplot.pack(side=TOP)
    self.text.pack(side=BOTTOM, expand=YES, fill=BOTH)
    
    #----------------------------------------------------------
    # Process
    #----------------------------------------------------------
    self.readFile()


#----------------------------------------------------------
# this is for plotting graphs
#----------------------------------------------------------
class PlotWindow:

  #----------------------------------------------------------
  def onExit(self):
    self.config.writeIniFile()
    #self.master.quit()
    #sys.exit(0)
    self.master.destroy()
    
  #----------------------------------------------------------
  def aboutBox(self):
    tkMessageBox.showinfo(self.config.name + " " + self.config.version,
      "Plotting 2D Graphs\nAuthor: Bernd Aberle\n\n" + HELP_TEXT)

  #----------------------------------------------------------
  def onSafePostscript(self):
##    sc = Pmw.ScrolledCanvas(root, borderframe=1, labelpos=N,
##                            label_text='ScrolledCanvas', usehullsize=1,
##                            hull_width=400,	hull_height=300)
##    img = PhotoImage(file='img52.gif')
##    sc.create_image(145,280, image=img, anchor=CENTER)
##    sc.pack()
##    sc.resizescrollregion()

    filename = tkFileDialog.asksaveasfilename(initialfile='PlotWindow.ps')
    if filename != None:
      self.canvas.postscript(file=filename)

  #----------------------------------------------------------
  def onRecentList(self, index):
    filename = self.popupmenu.entrycget(index + self.recent_offset, 'label')
    self.openFile(filename)
    
  #----------------------------------------------------------
  def setRecentFilesMenu(self):
    irecent = 0
    for recentfile in self.config.recentfiles:
      self.popupmenu.entryconfig(irecent+self.recent_offset, label=recentfile)
      irecent = irecent + 1

  #----------------------------------------------------------
  def onOpenFile(self):
    idir, ifile = self.config.filepath, 'test1.txt'
    # filename = tkFileDialog.askopenfilename(initialdir=idir, initialfile=ifile)
    filename = tkFileDialog.askopenfilename(initialdir=idir)
    self.openFile(filename)
    
  #----------------------------------------------------------
  def openFile(self, filename):
    if filename == "":
      return

    self.config.filepath = os.path.dirname(filename)
    asciifile = AsciiDataFile(filename)
    #asciifile.readData()
    analyser = AsciiDataFileAnalyser(filename, self, asciifile)
    
    self.filedata.append(asciifile)
    self.config.addRecentFile(filename)
    self.setRecentFilesMenu()

  #----------------------------------------------------------
  def plotFile(self, filename, startline=0, ixcolumn=0, iycolumn=0):
    self.config.filepath = os.path.dirname(filename)
    asciifile = AsciiDataFile(filename)
    asciifile.readData(startline, ixcolumn, iycolumn)
    analyser = AsciiDataFileAnalyser(filename, self, asciifile)
    
    self.filedata.append(asciifile)
    self.config.addRecentFile(filename)
    self.setRecentFilesMenu()
    self.plot()
    
  #----------------------------------------------------------
  # get minium and maximum value of all curves
  #
  def getMinMaxOfAllCurves(self):

    if len(self.filedata) >= 1:
      found = 0
      icurve = 0
      while found == 0 and icurve < len(self.filedata):
        try:
          self.xminplot = self.filedata[icurve].xminplot
          self.xmaxplot = self.filedata[icurve].xmaxplot
          self.yminplot = self.filedata[icurve].yminplot
          self.ymaxplot = self.filedata[icurve].ymaxplot
          found = 1
        except:
          continue
        icurve = icurve + 1

      if found == 0:
        return

      for icurve in range(len(self.filedata)):
        try:
          if self.filedata[icurve].xminplot < self.xminplot:
            self.xminplot = self.filedata[icurve].xminplot
          if self.filedata[icurve].yminplot < self.yminplot:
            self.yminplot = self.filedata[icurve].yminplot
          if self.filedata[icurve].xmaxplot > self.xmaxplot:
            self.xmaxplot = self.filedata[icurve].xmaxplot
          if self.filedata[icurve].ymaxplot > self.ymaxplot:
            self.ymaxplot = self.filedata[icurve].ymaxplot
        except:
          continue

  #----------------------------------------------------------
  # transforms an x value into screen coordinates for plotting
  #
  def getXScreenCoord(self, data, icurve=0, reverse=0):

    region = self.getPlotRegion(icurve)
    try:
      xfac = float(self.xsize) / (region['xmax'] - region['xmin'])
    except:
      return

    if reverse == 0:
      # get screen x coordinate value corresponding to data
      x1 = (float(data) - region['xmin']) * xfac + self.xoffset
    else:
      # get data value corresponding to screen x coordinate
      x1 = (data - self.xoffset) / xfac + region['xmin']
    return x1

  #----------------------------------------------------------
  # transforms an y value into screen coordinates for plotting
  #
  def getYScreenCoord(self, data, icurve=0, reverse=0):

    region = self.getPlotRegion(icurve)
    try:
      yfac = float(self.ysize) / (region['ymax'] - region['ymin'])
    except:
      return

    if reverse == 0:
      # get screen y coordinate value corresponding to data
      y1 = self.ysize - (float(data) - region['ymin']) * yfac + self.yoffset
    else:
      # get data value corresponding to screen y coordinate
      y1 = (self.ysize - (data - self.yoffset)) / yfac + region['ymin']
    return y1

  #----------------------------------------------------------
  # transforms a (x,y) pair into screen coordinates for plotting
  # if reverse is not 0, the reverse operation will be done
  #
  def getScreenCoord(self, data, icurve=0, reverse=0):
    return [self.getXScreenCoord(data[0], icurve, reverse), self.getYScreenCoord(data[1], icurve, reverse)]

  #----------------------------------------------------------
  # calculates the minima and maxima of values to plot
  #
  def getPlotRegion(self, icurve = 0):

    if self.ownyaxis.get() == 1:
      xmin = self.filedata[icurve].xminplot      
      xmax = self.filedata[icurve].xmaxplot      
      ymin = self.filedata[icurve].yminplot      
      ymax = self.filedata[icurve].ymaxplot      
    else:
      xmin = self.xminplot      
      xmax = self.xmaxplot      
      ymin = self.yminplot      
      ymax = self.ymaxplot
    region = {'xmin': xmin, 'ymin': ymin, 'xmax': xmax, 'ymax': ymax}
    return region

  #----------------------------------------------------------
  # plot all curves and annotation
  #
  def plot(self):

    self.clearWindow()
    if len(self.filedata) >= 1:
      for icurve in range(len(self.filedata)):
        if len(self.filedata[icurve].xaxis) < 2:
          continue      
        self.color = self.plotcolor[icurve % (len(self.plotcolor))]
        if self.ownyaxis.get() == 1:
          self.drawTicks(icurve) # draw ticks a little bit lower than previous
        else:
          self.getMinMaxOfAllCurves()
        self.plotcurve(icurve)

      if self.ownyaxis.get() == 0:
        self.drawTicks()

    for ianno in range(len(self.annotations)):
      anno = self.annotations[ianno]
      anno.textwidget = self.canvas.create_text(anno.x, anno.y, text=anno.text, anchor=W,
        fill=anno.color, font=tkFont.Font(size=anno.charsize))
      self.annotations[ianno].mybbox = self.canvas.bbox(anno.textwidget)
        
  #----------------------------------------------------------
  # plot the curve with index icurve
  #
  def plotcurve(self, icurve=0):

    if len(self.filedata[icurve].xaxis) < 2:
      return

    for idx in range (1, len(self.filedata[icurve].xaxis)):
      # calculate max and min values of region to plot
      region = self.getPlotRegion(icurve)
      xmin = region['xmin']      
      ymin = region['ymin']      
      xmax = region['xmax']      
      ymax = region['ymax']      

      # get start and end point for the line to draw
      tmp1 = [self.filedata[icurve].xaxis[idx - 1], self.filedata[icurve].yaxis[idx - 1]]
      tmp2 = [self.filedata[icurve].xaxis[idx], self.filedata[icurve].yaxis[idx]]

      if (tmp1[0] >= xmin and tmp1[0] <= xmax and tmp1[1] >= ymin and tmp1[1] <= ymax) or \
        (tmp2[0] >= xmin and tmp2[0] <= xmax and tmp2[1] >= ymin and tmp2[1] <= ymax):
        # convert the to points to screen coordinates
        d1 = self.getScreenCoord(tmp1, icurve)
        d2 = self.getScreenCoord(tmp2, icurve)
        try:
          self.canvas.create_line(d1[0], d1[1], d2[0], d2[1], fill=self.color)
        except:
          pass
        
  #----------------------------------------------------------
  # annotates x-and y-axis
  #
  def formatTicks(self, data):
    if abs(data) >= 10000 or abs(data) <= 1e-4:
      return fpformat.sci( data, 2)
    if abs(data) >= 10:
      return fpformat.fix( data, 2)
    else:
      return fpformat.fix( data, 3)
      
  #----------------------------------------------------------
  # annotates x-and y-axis
  #
  def drawTicks(self, icurve=0):    

    try:
      if self.ownyaxis.get() == 1:
        xmin = self.filedata[icurve].xminplot      
        xmax = self.filedata[icurve].xmaxplot      
        ymin = self.filedata[icurve].yminplot      
        ymax = self.filedata[icurve].ymaxplot      
      else:
        xmin = self.xminplot      
        xmax = self.xmaxplot      
        ymin = self.yminplot      
        ymax = self.ymaxplot      

      xticstep = float( fpformat.sci( (xmax - xmin) / 10.0, 1) )
      xtictmp = float( self.formatTicks(xmin) )
      xtic = []
      while xtictmp < xmax:
        xtic.append( self.formatTicks(xtictmp) )
        xtictmp = xtictmp + xticstep

      yticstep = float( fpformat.sci( (ymax - ymin) / 10.0, 1) )
      ytictmp = float( self.formatTicks(ymin) )
      #print ymax, ymin, yticstep, ytictmp
      ytic = []
      while ytictmp < ymax:
        ytic.append( self.formatTicks(ytictmp) )
        ytictmp = ytictmp + yticstep

      for idx in range(len(xtic)):
        x1 = self.getXScreenCoord(float(xtic[idx]), icurve)
        self.canvas.create_line(x1, self.ysize + self.yoffset, x1, self.ysize + self.yoffset - 8,
          fill=self.color)
        self.canvas.create_text(x1, self.ysize + self.yoffset + (icurve + 1) * 10,
          text=xtic[idx], fill=self.color)

      for idx in range(len(ytic)):
        y1 = self.getYScreenCoord(float(ytic[idx]), icurve)
        self.canvas.create_line(self.xoffset, y1, self.xoffset + 8, y1, fill=self.color)
        self.canvas.create_text(self.xoffset - 32, y1 + (icurve + 1) * 10 - 12, text=ytic[idx], fill=self.color)
    except:
      return
      
  #----------------------------------------------------------
  def resetData(self):
    self.filedata = []
    self.clearWindow()
    self.color = self.plotcolor[0]

  #----------------------------------------------------------
  # left click : remove the zoom rectangle, insert annotation
  #
  def onLeftClick(self, event):
    self.startmotion = 0
    self.startx = event.x
    self.starty = event.y
    self.popupmenu.unpost()

    if len(self.filedata) >= 1:
      sctmp = self.getScreenCoord([event.x, event.y], 0, -1)
      self.linfovar.set("clicked at " + fpformat.sci(sctmp[0], 3) + "  " + fpformat.sci(sctmp[1], 3))

    if self.rectanglethere != 0:
      self.canvas.delete(self.rectangle)
      self.rectanglethere = 0

    if self.annotation_mode == 1:
      # insert an annotation
      self.anno.x = event.x
      self.anno.y = event.y
      self.anno.textwidget = self.canvas.create_text(self.anno.x, self.anno.y,
        text=self.anno.text, anchor=W, fill=self.anno.color,
        font=tkFont.Font(size=self.anno.charsize))
      self.anno.mybbox = self.canvas.bbox(self.anno.textwidget)
      self.annotations.append(self.anno)
      self.annotation_mode = 0

    if self.annotation_mode == 2:
      # remove an annotation
      chi2 = 1e99
      idel = 999
      for ianno in range(len(self.annotations)):
        anno = self.annotations[ianno]
        chi2tmp = sqrt( \
          (float(event.x) - float(anno.mybbox[2]+anno.mybbox[0])/2.)**2 + \
          (float(event.y) - float(anno.y))**2 )
        if chi2tmp < chi2:
          chi2 = chi2tmp
          idel = ianno
      if idel < len(self.annotations) and chi2 < 50:
        del self.annotations[idel]
        self.plot()
      self.annotation_mode = 0
      
  #----------------------------------------------------------
  # open the context menu 
  #
  def onRightClick(self, event):
    self.popupmenu.post(event.x_root, event.y_root)
  
  #----------------------------------------------------------
  # zoom out
  #
  def zoomOutCompletely(self):
    if len(self.filedata) >= 1:
      for icurve in range(len(self.filedata)):
        if len(self.filedata[icurve].xaxis) < 2:
          continue
        # reset to maximum plot region
        self.filedata[icurve].xminplot = self.filedata[icurve].xmin
        self.filedata[icurve].xmaxplot = self.filedata[icurve].xmax
        self.filedata[icurve].yminplot = self.filedata[icurve].ymin
        self.filedata[icurve].ymaxplot = self.filedata[icurve].ymax
      self.plot()

  #----------------------------------------------------------
  # zoom out on x-axis
  #
  def onZoomOutX(self):
    if len(self.filedata) >= 1:
      for icurve in range(len(self.filedata)):
        if len(self.filedata[icurve].xaxis) < 2:
          continue
        xrange = self.filedata[icurve].xmaxplot - self.filedata[icurve].xminplot
        self.filedata[icurve].xminplot = self.filedata[icurve].xminplot - xrange/4.
        self.filedata[icurve].xmaxplot = self.filedata[icurve].xmaxplot + xrange/4.
      self.plot()
    
  #----------------------------------------------------------
  # zoom out on y-axis
  #
  def onZoomOutY(self):
    if len(self.filedata) >= 1:
      for icurve in range(len(self.filedata)):
        if len(self.filedata[icurve].xaxis) < 2:
          continue
        yrange = self.filedata[icurve].ymaxplot - self.filedata[icurve].yminplot
        self.filedata[icurve].yminplot = self.filedata[icurve].yminplot - yrange/4.
        self.filedata[icurve].ymaxplot = self.filedata[icurve].ymaxplot + yrange/4.
      self.plot()
    
  #----------------------------------------------------------
  # zoom out
  #
  def onLeftButtonDoubleClick(self, event):
    self.startmotion = 0
    self.zoomOutCompletely()
    
  #----------------------------------------------------------
  # drag the zoom rectangle
  #
  def onLeftClickMotion(self, event):
    self.startmotion = 1
    self.endx = event.x
    self.endy = event.y

    if self.rectanglethere != 0:
      self.canvas.delete(self.rectangle)
    self.rectangle = self.canvas.create_rectangle(self.startx, self.starty,
      self.endx, self.endy, outline="gray")
    self.rectanglethere = 1

  #----------------------------------------------------------
  def zoomIn(self, event):
    if self.startmotion == 1:
      self.startmotion = 0
      if len(self.filedata) >= 1:
        self.endx = event.x
        self.endy = event.y
        # plot region must be a rectangle not a line
        if self.startx == self.endx or self.starty == self.endy:
          return
        for icurve in range(len(self.filedata)):
          if len(self.filedata[icurve].xaxis) < 2:
            continue

          if self.ownyaxis.get() == 0:
            # if all curves use the same axis, take index 0
            icurvetmp = 0
          else:
            # if the curves use their own axis, take it's own index
            icurvetmp = icurve

          # get the curve values of the dragged rectangle
          xx = [self.getXScreenCoord(self.startx, icurvetmp, -1),
                self.getXScreenCoord(self.endx, icurvetmp, -1)]
          yy = [self.getYScreenCoord(self.starty, icurvetmp, -1),
                self.getYScreenCoord(self.endy, icurvetmp, -1)]
          xx.sort()
          yy.sort()
          self.filedata[icurve].xminplot = xx[0]
          self.filedata[icurve].xmaxplot = xx[1]
          self.filedata[icurve].yminplot = yy[0]
          self.filedata[icurve].ymaxplot = yy[1]

        self.plot()

  #----------------------------------------------------------
  # possibly the end of a dragging, so plot the region
  #
  def onLeftClickRelease(self, event):
    if self.startmotion == 1:
      self.zoomIn(event)
      self.startmotion = 0

  #----------------------------------------------------------
  def onYaxisCheck(self):
    self.zoomOutCompletely()

  #----------------------------------------------------------
  def onResizeWindow(self, event):
    self.windowxsize = event.width
    self.windowysize = event.height
    self.xsize = event.width * 0.8
    self.ysize = event.height * 0.8
    self.xoffset = max(70, self.xsize / 10)
    self.yoffset = self.ysize / 10
    self.plot()

  #----------------------------------------------------------
  def clearWindow(self):
    self.startx = self.xoffset
    self.starty = self.yoffset
    self.endx = self.xoffset + self.xsize
    self.endy = self.yoffset + self.ysize

    self.canvas.create_rectangle(0, 0, self.windowxsize,
      self.windowysize, fill="white", outline="white")
    self.canvas.create_rectangle(self.xoffset, self.yoffset,
      self.xoffset + self.xsize, self.yoffset + self.ysize, fill="white")
  
  #----------------------------------------------------------
  # insert an annotation string
  #
  def onNewAnnotation(self):
    dialog = AnnotationDialog(self.master)
    self.anno = dialog.result
    if self.anno != None:
      # print self.anno.text, self.anno.charsize, self.anno.color, self.anno.orientation
      self.annotation_mode = 1
      
  #----------------------------------------------------------
  # delete an annotation string
  #
  def onDeleteAnnotation(self):
    tkMessageBox.showinfo("Deleting annotation",
      "Click with left mouse button on \nthe annotation you want to delete")
    self.annotation_mode = 2
  
  #----------------------------------------------------------
  def __init__(self, master):

    self.master = master
    self.plotcolor = ["black", "red", "blue", "magenta", "darkblue", "darkgreen", "brown"]
    self.color = self.plotcolor[0]
    self.filedata = []
    self.annotations = []
    self.anno = None
    self.config = MyConfig()
    self.annotation_mode = 0
    self.master.title(self.config.name + " " + self.config.version)
    self.xminplot = 0.0
    self.xmaxplot = 1.0
    self.yminplot = 0.0
    self.ymaxplot = 1.0
    
    #----------------------------------------------------------
    # Menu definition
    #----------------------------------------------------------
    menubar = Menu(self.master)
    self.master.config(menu=menubar)

    self.filemenu = Menu(menubar)
    menubar.add_cascade(label="File", menu=self.filemenu)
    self.filemenu.add_command(label="Open...", command=self.onOpenFile)
    self.filemenu.add_command(label="Save as postscript...", command=self.onSafePostscript)
    self.filemenu.add_command(label="Exit", command=self.onExit)

    helpmenu = Menu(menubar)
    menubar.add_cascade(label="Help", menu=helpmenu)
    helpmenu.add_command(label="Usage...", command=self.aboutBox)
    
    #----------------------------------------------------------
    # Canvas
    #----------------------------------------------------------
    self.windowxsize = 800
    self.windowysize = 600
    self.xsize = 600
    self.ysize = 400
    self.xoffset = self.xsize / 10
    self.yoffset = self.ysize / 10

    self.startx = self.xoffset
    self.starty = self.yoffset
    self.endx = self.xoffset + self.xsize
    self.endy = self.yoffset + self.ysize

    self.rectanglethere = 0
    self.startmotion = 0

    self.canvas = Canvas(self.master, width = self.xoffset + self.xsize + 30,
      height = 2 * self.yoffset + self.ysize, bg="white")

    self.canvas.create_rectangle(self.xoffset, self.yoffset,
      self.xoffset + self.xsize, self.yoffset + self.ysize, fill="white")

    #----------------------------------------------------------
    # Popup menu
    #----------------------------------------------------------
    self.popupmenu = Menu(self.master, tearoff=0)
    self.annotationmenu = Menu(self.popupmenu)
    self.popupmenu.add_cascade(label="Annotation", menu=self.annotationmenu)
    self.annotationmenu.add_command(label="Create new...", command=self.onNewAnnotation)
    self.annotationmenu.add_command(label="Delete...", command=self.onDeleteAnnotation)
    self.popupmenu.add_command(label="Zoom out (x-axis)", command=self.onZoomOutX)
    self.popupmenu.add_command(label="Zoom out (y-axis)", command=self.onZoomOutY)
    self.popupmenu.add_separator()    
    self.recent_offset = 4 # this is the index of the first recent-file in the popup menu

    self.popupmenu.add_command(label=" ", command=lambda:self.onRecentList(0))
    self.popupmenu.add_command(label=" ", command=lambda:self.onRecentList(1))
    self.popupmenu.add_command(label=" ", command=lambda:self.onRecentList(2))
    self.popupmenu.add_command(label=" ", command=lambda:self.onRecentList(3))
    self.popupmenu.add_command(label=" ", command=lambda:self.onRecentList(4))
    self.popupmenu.add_command(label=" ", command=lambda:self.onRecentList(5))

    #----------------------------------------------------------
    # Other widgets
    #----------------------------------------------------------
    bframe = Frame(self.master)
    bclear = Button(bframe, text="Clear",  fg="blue", command=self.clearWindow)
    bplot = Button(bframe, text="Plot",  fg="blue", command=self.plot)
    self.ownyaxis = IntVar()
    cyaxis = Checkbutton(bframe, text="Scale each plot to it's own axis",  fg="blue",
      variable=self.ownyaxis, command=self.onYaxisCheck)
    self.linfovar = StringVar()
    linfo = Label(bframe, textvariable=self.linfovar)

    cyaxis.grid(column=0, row=0, sticky=W)
    bclear.grid(column=1, row=0, sticky=W)
    bplot.grid(column=2, row=0, sticky=W)
    linfo.grid(column=3, row=0, sticky=W)
    
    #----------------------------------------------------------
    # Event bindings
    #----------------------------------------------------------
    self.canvas.bind("<Button-1>", self.onLeftClick)
    self.canvas.bind("<Double-Button-1>", self.onLeftButtonDoubleClick)
    self.canvas.bind("<Button-3>", self.onRightClick)
    self.canvas.bind("<B1-Motion>", self.onLeftClickMotion)
    self.canvas.bind("<B1-ButtonRelease>", self.onLeftClickRelease)
    self.canvas.bind("<Configure>", self.onResizeWindow)
    self.master.protocol("WM_DELETE_WINDOW", self.onExit)
		    
    #----------------------------------------------------------
    # initialise
    #----------------------------------------------------------
    self.config.readIniFile()
    self.setRecentFilesMenu()
    
    #----------------------------------------------------------
    # Layout
    #----------------------------------------------------------
    # self.listboxgrid(column=0, row=0, sticky=W)
    
    #bclear.grid(column=0, row=0, sticky=W)
    #bplot.grid(column=1, row=0, sticky=W)
    #self.canvas.grid(column=0, row=1, sticky=W, rowspan=1)

    bframe.pack(side=TOP)
    self.canvas.pack(expand=1, side=BOTTOM, fill=BOTH)


#----------------------------------------------------------
# main routine
#----------------------------------------------------------
print '---------------------------------------------------'
print "running : ", sys.argv

filename = ''
if len(sys.argv) >= 2:
  filename = sys.argv[1]

startline = 0
if len(sys.argv) >= 3:
  startline = int(sys.argv[2])

ixcolumn = 0
if len(sys.argv) >= 4:
  ixcolumn = int(sys.argv[3])

iycolumn = 0
if len(sys.argv) >= 5:
  iycolumn = int(sys.argv[4])

root = Tk()
plotWindow = PlotWindow(root)
if filename != '':
  plotWindow.plotFile(filename, startline, ixcolumn, iycolumn)

root.mainloop()
