#!/usr/local/bin/python
"""
    State Dump Utility -- somewhat like a core dump for Python

    
-----------------------------------------------------------------------------
(c) Copyright by Marc-Andre Lemburg (mailto:mal@lemburg.com)

    Permission to use, copy, modify, and distribute this software and its
    documentation for any purpose and without fee or royalty is hereby granted,
    provided that the above copyright notice appear in all copies and that
    both that copyright notice and this permission notice appear in
    supporting documentation or portions thereof, including modifications,
    that you make.

    THE AUTHOR MARC-ANDRE LEMBURG DISCLAIMS ALL WARRANTIES WITH REGARD TO
    THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
    FITNESS, IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL,
    INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
    FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
    NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
    WITH THE USE OR PERFORMANCE OF THIS SOFTWARE !


"""
__version__ = '0.2'

__package_info__ = """
BEGIN PYTHON-PACKAGE-INFO 1.0
Title:                  State dump utility
Description:            Dump the current state of the interpreter into a file
Current-Version:        0.2
Home-Page:              http://starship.python.net/~lemburg
Primary-Site:           http://starship.python.net/~lemburg/statedump.py
END PYTHON-PACKAGE-INFO
"""

import sys,types

class StateDump:

    """ output the current state of the python interpreter to
        a file, starting from an object in its namespace

        * can also be installed as exit-function to help in
          debugging code (local variables are printed after
          the traceback)
        * any file with a write-method is accepted
        * an instance of this class can only be used once 
          for dumping a state !!!
    """

    def reset(self):
        # these modules won't be recursed into
        self.Cache = {
            '__builtins__':1,
            '__builtin__':1,
            'regex':1,
            'regsub':1,
            'linecache':1,
            'os':1,
            'path':1,
            'posixpath':1,
            'stat':1,
            'urllib':1,
            'UserDict':1,
            'UserList':1,
            'operator':1,
            'traceback':1,
            'string':1,
            'signal':1,
            'strop':1,
            'types':1,
            'socket':1
            }
        # never show Cache
        self.Cache[id(self.Cache)] = 2
        # these types won't show
        self.OmitTheseTypes = (types.BuiltinFunctionType,)
        # omit special methods/vars (__***__) ?
        self.OmitSpecials = 1
        # indentation
        self.indent_per_level = 4
        # show globals on error ?
        self.OmitGlobals = 1

    def __init__(self,file=sys.stderr):

        # reset
        self.reset()
        # set output file
        self.file = file

    def dump(self,object,indent=0,title=None):

        """ dump object recursively
            * object can be a namespace dictionary or an object having
              a __dict__-Dictionary
            * file must be open for writing
            * title 
            * take care to call reset before using this method a second time
              (because if not, only the new items will be shown)
        """

        import string,time
        if title == None:
            title = 'Python State Dump from "%s" at %s'% \
                    (string.join(sys.argv),time.ctime(time.time()))
        self.file.write('\n%s\n%s\n'%(title,len(title)*'-'))
        self.dump_notitle(object,indent)

    def dump_main(self):

        """ dump from top level """

        import __main__
        self.dump(__main__)

    def dump_on_error(self):

        """ dump traceback in case of an unhandled exception """

        if hasattr(sys,'last_traceback'):
            tb = sys.last_traceback
            while tb.tb_next != None: tb = tb.tb_next
            frame = tb.tb_frame
            self.dump(frame.f_locals,title='Traceback: Local Namespace')
            if not self.OmitGlobals:
                self.dump(frame.f_globals,title='Traceback: Global Namespace')

    def install(self):

        """ install the state dump utility as system exit function """

        sys.exitfunc = self.dump_on_error

    def dump_notitle(self,object,indent=0):

        """ worker: same as dump(), but without titles """

        if self.Cache.has_key(id(object)):
            c =  self.Cache[id(object)]
            if c == 1:
                self.file.write('...already shown\n')
            elif c == 2:
                self.file.write('...omitted\n')
            else:
                self.file.write('...cached with code %i\n'%c)
            return

        objtype = type(object)

        if objtype == types.ModuleType:
            module_name = object.__name__
            if self.Cache.has_key(module_name):  
                self.file.write('...omitted\n')
                return
            self.Cache[module_name] = 1

        if hasattr(object,'__dict__'):
            items = object.__dict__.items()
            self.Cache[id(object)] = 1
        elif hasattr(object,'items'):
            items = object.items()
            self.Cache[id(object)] = 1
        else:
            self.file.write(repr(object)+'\n')
            return

        if len(items) > 0:
            items.sort()
            try:
                self.file.write('(%s of %s at %x)...\n'%(type(object),repr(object.__class__),id(object)))
            except:
                self.file.write('(%s at %x)...\n'%(type(object),id(object)))
        else:
            self.file.write(repr(object)+'\n')
            return

        for key,data in items:
            if type(data) in self.OmitTheseTypes: continue
            if self.OmitSpecials and type(key) == types.StringType \
               and (key[:2] == key[-2:] == '__'): continue
            self.file.write(indent*' '+str(key)+': ')
            self.dump_notitle(data,indent+self.indent_per_level)

def test(a,b):

    print '*** This is the current state...'
    c = 4
    sd.dump_main()
    print
    print '*** Producing an unhandled exception...'
    print a/b

if __name__ == '__main__':
    global sd
    sd = StateDump(sys.stdout)
    sd.install()
    a = 2
    b = 0
    test(a,b)

### uncomment to have the import of this module install itself on load
#
#StateDump().install()

